### Persefone.jl - a model of agricultural landscapes and ecosystems in Europe.
###
### This file is responsible for managing the animal modules and contains functions
### and types integrating the nature submodel with the rest of Persefone.
###


## An enum used to assign a sex to each animal
@enum Sex hermaphrodite male female
#XXX hermaphrodite probably is not needed and its implementation is iffy

"""
    Animal

This is the generic agent type for all animals. Species are differentiated
by trait dictionaries passed by them during initialisation. (Note that each
trait variable can still be accessed as if it were a normal field name,
i.e. the trait `phase` can be accessed and modified with `animal.phase`.)
"""
@agent Animal GridAgent{2} begin
    #TODO remove Agents.jl-related code, reimplement this more cleanly (#72)
    #XXX is it (performance-)wise to use a dict for the traits?
    # Doesn't that rather obviate the point of having an agent struct?
    # If I could move the mutable traits to the struct, I wouldn't need
    # to deepcopy the speciesdict.
    #XXX add `phase` and `species` (rather than `name`) directly to the struct?
    traits::Dict{String,Any}
    energy::Union{EnergyBudget,Nothing} # DEB is optional
    parents::Tuple{Int64,Int64} #XXX two parents for hermaphrodites?
    sex::Sex
    age::Int
end

# DEB is optional
Animal(id::Int64, pos::Tuple{Int64,Int64}, traits::Dict{String,Any},
       parents::Tuple{Int64,Int64}, sex::Sex, age::Int) =
           Animal(id, pos, traits, nothing, parents, sex, age)

# Overloading the `getproperty` and `setproperty!` methods for `Animal` allows
# us to write `animal.property` instead of `animal.traits["property"]`.
# (This is inspired by Agents.jl in `model.jl`.)

function Base.getproperty(animal::Animal, s::Symbol)
    if s in fieldnames(Animal)
        return getfield(animal, s)
    else
        return animal.traits[String(s)]
    end
end

function Base.setproperty!(animal::Animal, s::Symbol, x)
    if s in fieldnames(Animal)
        setfield!(animal, s, x)
    else
        animal.traits[String(s)] = x
    end
end

    
"""
    animalid(animal)

A small utility function to return a string with the species name and ID of an animal.
"""
function animalid(a::Animal)
    return "$(a.name) $(a.id)"
end

"""
    stepagent!(animal, model)

Update an animal by one day, executing it's currently active phase function.
"""
function stepagent!(animal::Animal, model::AgentBasedModel)
    animal.age += 1
    animal.traits[animal.phase](animal,model) #FIXME -> note to self: why?
end

"""
    initnature!(model)

Initialise the model with all simulated animal populations.
"""
function initnature!(model::AgentBasedModel)
    # The config file determines which species are simulated in this run
    for speciesname in @param(nature.targetspecies)
        species = @eval $(Symbol(speciesname))($model)
        species["initialise!"](species, model)
    end
    # Initialise the data output
    initecologicaldata(model)
end

"""
    updatenature!(model)

Run processes that affect all animals.
"""
function updatenature!(model::AgentBasedModel)
    # The migrant pool is sorted by date of return, so we can simply look at the top
    # of the stack to check whether any animals are returning today.
    while !isempty(model.migrants) && model.migrants[1].second <= model.date
        add_agent_pos!(model.migrants[1].first, model)
        @debug "$(animalid(model.migrants[1].first)) has returned."
        deleteat!(model.migrants, 1)
    end
    #XXX what else needs to go here?
end

#TODO test migration
